#!/usr/bin/env python

# b2u
b2u = open('uao250-b2u.big5.txt', 'r').readlines()
b2u = [line.strip().split(' ')
       for line in b2u
       if line.strip().startswith('0x')]
b2u = dict((int(b, 0), int(u, 0)) for (b, u) in b2u)

print """#include <stdint.h>
extern const uint16_t const b2u_table[];
extern const uint16_t const u2b_table[];
extern const uint8_t const b2u_ambiguous_width[];
"""
print "const uint16_t const b2u_table[0x10000] = {"
for i in range(0x10000):
    print '0x%04x,' % (i if i not in b2u else b2u[i]),
    if i % 10 == 9:
        print ''
print "};\n"

# u2b
u2b = open('uao250-u2b.big5.txt', 'r').readlines()
u2b = [line.strip().split(' ')
       for line in u2b
       if line.strip().startswith('0x')]
u2b = dict((int(u, 0), int(b, 0)) for (b, u) in u2b)

print "const uint16_t const u2b_table[0x10000] = {"
for i in range(0x10000):
    print '0x%04x,' % (i if i not in u2b else u2b[i]),
    if i % 10 == 9:
        print ''
print "};\n"

# ambiguous cjk width
ambcjk_data = open("ambcjk.big5.txt").readlines()
ambcjk = []
for entry in ambcjk_data:
    (a, b) = entry.strip().split()
    a = int(a, 0)
    b = int(b, 0)
    ambcjk = ambcjk + range(a, b+1)
print "const uint8_t const b2u_ambiguous_width[0x10000] = {"
for i in range(0, 0x10000):
    if i in b2u and b2u[i] in ambcjk:
        print "1,",
    else:
        print "0,",
    if i % 25 == 24:
        print ''

print "};\n"



#!/usr/bin/env python
#
# parse register.log and create reglog.db
#
# please first test and fix your register.log, then 
# try writing to database.
# 
# Create: piaip

# python 2.5:
#import sqlite3
# python 2.4 or prior:
from pysqlite2 import dbapi2 as sqlite3

import sys
import time

def initdb(dbfn) :
    db = sqlite3.connect(dbfn)
    c = db.cursor()
    # separate the tables to 'approved' and 'rejected',
    # since the 'rejected' is rarely used...
    c.execute('create table if not exists approved ' +
            '(uid text, name text, career text, ' +
            ' addr text, phone text, date integer, approved text)');
    c.execute('create table if not exists rejected ' +
            '(uid text, name text, career text, ' +
            ' addr text, phone text, date integer, rejected text)');
    return (db, c)

def processblk(m) :
    #print "process blk!"
    if not m.has_key('uid') :
        print "no uid record at line: ", lineno
        sys.exit(-1)
    if (not m.has_key('approved')) and (not m.has_key('rejected')) :
        print "invalid record (action?) at line: ", lineno
        sys.exit(-1)
    if c != None :
        type = 'rejected';
        if m.has_key('approved') : type = 'approved';
        # write into database
        sql = 'insert into ' + type + ' values (?,?,?,?,?,?,?)';
        # follow the orders
        c.execute(sql, (m['uid'], m['name'], m['career'], m['addr'], m['phone'], m['date'], m[type]));

# default input
fn = "register.log" 
dbfn= ""
db = None
c = None
if len (sys.argv) < 1 :
    print "syntax:", sys.argv[0], "register.log [reglog.db]"
    sys.exit(0)
if len (sys.argv) > 1 :
    fn = sys.argv[1]
if len (sys.argv) > 2 :
    dbfn = sys.argv[2]

print "Opening ", fn, "..."
f = open(fn, "rt")

if dbfn != '' :
    (db, c) = initdb(dbfn)

m={}
lineno = 0
tickets = 0
for s in f :
    lineno = lineno +1
    # for each line...
    if s.startswith("num:")   : continue
    if s.startswith("email:") : continue
    if s.startswith("ident:") : continue
    if s.startswith("]") : continue
    if s.startswith("----")   :
            # ready to build new one
            processblk(m)
            m = {}
            tickets = tickets +1
            if tickets % 10000 == 0 :
                if db : db.commit()
                print tickets, "forms processed."
            continue
    v = s.split(': ', 1)
    if len(v) != 2 :
        print "error at line ", lineno
        sys.exit(-1)
    k = v[0].lower()
    if (m.has_key(k)) :
        print "key duplicated at line ", lineno
        sys.exit(-1)
    # convert timestamp
    v = v[1].strip()
    if (k == 'date') :
        # strip final weekday
        v = v.split(' ') # mdy HMS abrv
        if len(v) < 2 or len(v) > 3 :
            print "invalid date entry at line ", lineno
            sys.exit(-1)
        v = v[0] + ' ' + v[1]
        # build timestamp
        v = int(time.mktime(time.strptime(v, "%m/%d/%Y %H:%M:%S")))
    m[k] = v

f.close()

# vim:sw=4:expandtab



#!/usr/bin/env python
#-*- coding: big5 -*-

import collections
import math
import os
import time

BBSHOME = '/home/bbs'
INPUT_FILE = '%s/log/angel_perf.txt' % (BBSHOME)
# A regular week = 590 samples.
SAMPLE_MINIMAL = 200

Entry = collections.namedtuple('Entry', 'sample pause1 pause2 max avg std')
DEBUG = None
REPORT_SAMPLE_STAT = False

PREFIX_DOC = '''== 本週小天使查翅膀建議名單 (系統自動產生: %s) ==

說明: 天使公會每十分鐘會統計一次所有的小天使狀態，
      確認 (1) 小天使當時是否在線上 (2) 神諭呼叫器是否停收/關閉。
      依據天使之書對於查翅膀的定義「抽查到太多次都沒開啟呼叫器」，
      若上線時間過少或是關呼叫時間過高則會出現在每週統計的名單中，
      以供品管及大天使參考。

  ** 為幫助理解名單產生方法及數字意義，簡單例子如下: 若有小天使 ABC 三名,
     A 在線上開呼叫, B 在線上但關閉呼叫器, C 不在線上:
     某日早上 10:30 系統進行統計，系統就會記錄:
                     [ID] [線上] [停收] [關閉]
                     A    1      0      0     
                     B    1      0      1     
                     C    0      0      0     
              10:31 使用者與小天使A在互丟水球 (丟水球不統計)
              10:32 C 上線了 (還不統計 - 十分鐘才算一次)
              10:33 A 改為停收 (還不統計 - 十分鐘才算一次)
              10:40 系統再次進行統計，更改記錄如下:
	            (ABC都在線上，A停收，B關閉，C開呼叫)
		     [ID] [線上] [停收] [關閉]
		     A    1+1=2  0+1=1  0+0=0  
		     B    1+1=2  0+0=0  1+1=2 
		     C    0+1=1  0+0=0  0+0=0  
		    且下次統計時間為 10:50。

     換句話說，(數字/6) 就大約是線上/停收/關閉的總小時數。
''' % (time.ctime())

def is_lazy(e):
    # 'LAZY'
    '\033[1;33m以下是[線上]統計結果過低(小於5)的小天使\033[m'
    # return e.sample < (e.avg - 1.0 * e.std)
    return e.sample < 5

def is_all_reject2(e):
    # 'ALL_REJECT2'
    '\033[1;31m以下是關呼叫統計比例過高([線上]減[關閉]與[停收]小於5)的小天使\033[m'
    return (e.pause2 + e.pause1 >= e.sample - 4)

def parse_perf_file(filename):
    data = {}
    max_sample = 0
    sum_sample = 0
    sum_sample_square = 0
    with open(filename, 'r') as f:
	for l in f:
	    ls = l.strip()
	    if ls.startswith('#') or ls == '':
		continue
            # format: no. uid sample pause1 pause
	    no, uid, sample, pause1, pause2 = ls.split()
	    data[uid] = map(int, (sample, pause1, pause2))
	    sample = int(sample)
	    sum_sample += sample
	    sum_sample_square += sample * sample
	    if sample > max_sample:
		max_sample = sample
    N = len(data) or 1
    avg_sample = sum_sample / N
    std_sample = math.sqrt((sum_sample_square -
			    (N * avg_sample * avg_sample)) / N)
    return max_sample, avg_sample, std_sample, data

def get_nick(uid):
    fn = '%s/home/%c/%s/angelmsg' % (BBSHOME, uid[0], uid)
    nick = ''
    if os.path.exists(fn):
	nick = open(fn).readline().strip().decode('big5').strip('%%[')
    else:
	nick = uid
    return (nick + '小天使'.decode('big5')).encode('big5')

def build_badges(max_sample, avg_sample, std_sample, data):
    result = {}
    filters = [is_all_reject2]
    for uid, e in data.items():
	nick = '%s (線上%d 停收%d 關閉%d)' % (get_nick(uid), e[0], e[1], e[2])
	if DEBUG:
	    nick += ' {%s/%d/%d/%d}' % (uid, e[0], e[1], e[2])
	entry = Entry(e[0], e[1], e[2], max_sample, avg_sample, std_sample)
	if is_lazy(entry):
	    badges = [is_lazy.__doc__]
	else:
	    badges = [f.__doc__ for f in filters if f(entry)]
	for b in badges:
	    if b not in result:
		result[b] = []
	    result[b] += [nick]
    return result

def main():
    max_sample, avg_sample, std_sample, data = parse_perf_file(INPUT_FILE)
    # print 'max=%f, avg=%f, std=%f' % (max_sample, avg_sample, std_sample)
    if max_sample < SAMPLE_MINIMAL:
    	exit()

    print PREFIX_DOC
    if DEBUG:
	print 'max, avg, std: %d, %d, %d' % (max_sample, avg_sample, std_sample)
    elif REPORT_SAMPLE_STAT:
	print ' SAMPLE 數最大值 / 平均 / 標準差: %d / %d / %d\n' % (
	       max_sample, avg_sample, std_sample)
    else:
	pass
    result = build_badges(max_sample, avg_sample, std_sample, data)
    for k, v in result.items():
	print '%s:\n  %s\n' % (k, '\n  '.join(v))

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright (c) 2012 Hung-Te Lin <piaip@csie.ntu.edu.tw>. All rights reserved.
# Use of this source code is governed by a BSD license.

"""
banipd: Daemon to check if an IP is banned, and sending to reason.
"""

import eventlet
import getopt
import logging
import pprint
import re
import signal
import string
import struct
import sys


_SERVER_ADDR = '127.0.0.1'
_SERVER_PORT = 5134
_CONFIG_FILE = "/home/bbs/etc/banip.conf"

g_tbl = {}


def check_ipv4(text):
    ip = text.split('.')
    if len(ip) != 4:
        raise ValueError('Not valid IPV4: %s' % text)
    for i in ip:
        ii = int(i)
        if ii < 0 or ii > 255:
            raise ValueError('Not valid IPV4: %s' % text)
    return True


def LoadConfigTable(filename):
    table = {}

    class Entry(object):
        def __init__(self):
            self.clear()

        def clear(self):
            self.ip = ''
            self.text = ''

        def complete(self):
            ips = self.build_ip()
            for ip in ips:
                if ip in table:
                    raise ValueError('duplicated IP: %s' % ip)
                table[ip] = self.text.rstrip() + "\n"
            self.clear()

        def add_ip(self, s):
            if '#' in s:
                s = s.partition('#')[0]
            self.ip += ' ' + s

        def add_text(self, s):
            self.text += s

        def build_ip(self):
            ips = self.ip.split()
            if not all((check_ipv4(ip) for ip in ips)):
                raise Exception('Bad IP.')
            return ips

    entry = Entry()
    with open(filename, 'rt') as f:
        for line in f:
            t = line.strip()
            if t.startswith('#'):
                continue
            if t and t[0] in string.digits:
                # create new entry, or continue
                if entry.text:
                    entry.complete()
                entry.add_ip(t)
            elif entry.ip:
                # more text
                entry.add_text(line)
            elif not t:
                continue
            else:
                raise ValueError("Text before IP definition: %s" % t)
    if entry.ip:
        entry.complete()
    return table


def handle_request(sock, fd):
    # Input: IP\n
    # Output: int32_t len, BYTE[len]
    fmt_len = '@i'
    try:
        ip = fd.readline().strip()
        msg = g_tbl.get(ip, '')
        logging.debug('Query [%s]: %s', ip, '[BANNED]' if msg else '[PASS]')
        fd.write(struct.pack(fmt_len, len(msg)))
        fd.write(msg)
    except:
        logging.exception("handle_request")
    finally:
        try:
            fd.close()
            sock.close()
        except:
            pass


def main(myname, argv):
    global g_tbl

    def handle_hup(signum, stack):
        logging.warn("Reload configuration table: %s" % config_file)
        try:
            newtbl = LoadConfigTable(config_file)
            g_tbl.clear()
            g_tbl.update()
            logging.warn("Successfully updated table.")
        except:
            logging.exception("Failed loading new config: %s" % config_file)
            return

    def usage():
        print ("Usage: %s [options] config_file\n"
               "-c: check only.\n"
               "-d: enable debug output.\n"
               "config_file: default to %s" % _CONFIG_FILE)
    try:
        opts, args = getopt.getopt(argv, "cd")
    except getopt.GetoptError, err:
        print str(err)
        usage()
        exit(2)

    level = logging.WARNING
    check_only = False
    test = None
    for o, a in opts:
        if o == '-c':
            check_only = True
        elif o == '-d':
            level = logging.DEBUG
        else:
            assert False, "Unkown param: %s" % o
    if len(args) > 1:
        print "Too many config files: %s" % args
        usage()
        exit(2)
    config_file = args[0] if args else _CONFIG_FILE
    logging.basicConfig(level=level, format='%(asctime)-15s %(message)s')
    logging.info('Checking config file: %s', config_file)
    g_tbl.clear()
    g_tbl.update(LoadConfigTable(config_file))
    if check_only:
        # pprint.pprint(g_tbl)
        print "%s: PASSED." % config_file
        return
    logging.warn("Serving at %s:%s [config:%s]...", _SERVER_ADDR, _SERVER_PORT,
                 config_file)
    signal.signal(signal.SIGHUP, handle_hup)
    server = eventlet.listen((_SERVER_ADDR, _SERVER_PORT))
    pool = eventlet.GreenPool()
    while True:
        try:
            new_sock, address = server.accept()
            pool.spawn_n(handle_request, new_sock, new_sock.makefile('rw'))
        except (SystemExit, KeyboardInterrupt):
            break


if __name__ == '__main__':
    main(sys.argv[0], sys.argv[1:])



#!/usr/bin/env python
# Usage: ./big5_gen.py > big5_tbl.py

import os

COMMON_SYS = '../../common/sys'
B2U_FILE = os.path.join(COMMON_SYS, 'uao250-b2u.big5.txt')
U2B_FILE = os.path.join(COMMON_SYS, 'uao250-u2b.big5.txt')

# b2u
b2u = open(B2U_FILE, 'r').readlines()
b2u = [line.strip().split(' ')
       for line in b2u
       if line.strip().startswith('0x')]
b2u = dict((int(b, 0), int(u, 0)) for (b, u) in b2u)

print """#!/usr/bin/env python

"""
print "b2u_table = ("
for i in range(0x10000):
    print '0x%04x,' % (i if i not in b2u else b2u[i]),
    if i % 10 == 9:
        print ''
print ")\n"

# u2b
u2b = open(U2B_FILE, 'r').readlines()
u2b = [line.strip().split(' ')
       for line in u2b
       if line.strip().startswith('0x')]
u2b = dict((int(u, 0), int(b, 0)) for (b, u) in u2b)

print "u2b_table = ("
for i in range(0x10000):
    print '0x%04x,' % (i if i not in u2b else u2b[i]),
    if i % 10 == 9:
        print ''
print ")\n"



#!/usr/bin/env python

import big5_tbl

def decode(s, strip_zero=True):
    ret = u''
    b = 0
    for i in s:
	if b:
	    b = (b << 8) | (ord(i))
	else:
	    b = ord(i)
	    if b >= 0x80:
		continue
	if (b == 0) and strip_zero:
	    break
	ret += unichr(big5_tbl.b2u_table[b])
	b = 0
    return ret

def encode(u):
    ret = ''
    for i in u:
	c = big5_tbl.u2b_table[ord(i)]
	b0 = (c) & 0xff
	b1 = (c >> 8) & 0xff
	if b1:
	    ret = ret + chr(b1)
	ret = ret + chr(b0)
    return ret



#!/usr/bin/env python

import struct
import collections
import sys
import logging
import os

from gevent import monkey; monkey.patch_all()
import gevent
import gevent.server
import leveldb

# Ref: ../../include/pttstruct.h
IDLEN = 12
FNLEN = 28
COMMENTLEN = 80


# Ref: ../../include/daemons.h
RequestFormatString = 'HH';
Request = collections.namedtuple('Request', 'cb operation')
CommentFormatString = "IIII%ds%ds" % (IDLEN + 1, COMMENTLEN + 1)
Comment = collections.namedtuple('Comment',
	'time ipv4 userref type userid msg')
CommentKeyFormatString = '%ds%ds' % (IDLEN + 1, FNLEN + 1)
CommentKey = collections.namedtuple('CommentKey', 'board file')
REQ_ADD = 1
REQ_QUERY = 2
_SERVER_ADDR = '127.0.0.1'
_SERVER_PORT = 5134
_DB_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)),
			'db_comments.db')


def UnpackComment(blob):
    def strip_if_string(v):
	return v.strip(chr(0)) if type(v) == str else v
    data = struct.unpack(CommentFormatString, blob)
    logging.debug("UnpackComment: %r" % (data,))
    return Comment._make(map(strip_if_string, data))

def UnpackCommentKey(blob):
    def strip_if_string(v):
	return v.strip(chr(0)) if type(v) == str else v
    data = struct.unpack(CommentKeyFormatString, blob)
    logging.debug("UnpackCommentKey: %r" % (data,))
    return CommentKey._make(map(strip_if_string, data))

def PackComment(comment):
    return struct.pack(CommentFormatString, *comment)

def LoadComment(key):
    blob = g_db.get(key)
    if blob is None:
	return ''
    else:
	return UnpackComment(g_db.get(key))

def SaveComment(keypak, comment):
    logging.debug("SaveComment: %r => %r" % (keypak, comment))
    key = '%s/%s' % (keypak.board, keypak.file)
    num = g_db.get(key)
    num = 1 if (num is None) else (int(num) + 1)
    g_db.set(key, '%d' % num)
    key += '#%08d' % (num)
    g_db.set(key, PackComment(comment))
    logging.info('Saved comment: %s', key)

def open_database(db_path):
    global g_db

    class LevelDBWrapper(object):
	def __init__(self, db):
	    self.db = db

	def get(self, key):
	    try:
		return self.db.Get(key)
	    except KeyError:
		return None

	def set(self, key, value):
	    self.db.Put(key, value)

    g_db = LevelDBWrapper(leveldb.LevelDB(db_path))
    return g_db

def handle_request(socket, _):
    fd = socket.makefile('rw')
    fmt_len = '@i'
    try:
        req = fd.read(struct.calcsize(RequestFormatString))
	req = Request._make(struct.unpack(RequestFormatString, req))
	logging.debug('Found request: %d' % req.operation)
        # TODO check req.cb
	if req.operation == REQ_ADD:
	    blob = fd.read(struct.calcsize(CommentFormatString))
	    keyblob = fd.read(struct.calcsize(CommentKeyFormatString))
	    SaveComment(UnpackCommentKey(keyblob), UnpackComment(blob))
	elif req.operation == REQ_QUERY:
	    raise NotImplementedError('REQ_QUERY is not implemented')
	else:
	    raise ValueError('Unkown operation %d' % req.operation)
    except:
        logging.exception("handle_request")
    finally:
        try:
            fd.close()
            socket.close()
        except:
            pass

def main(myname, argv):
    level = logging.INFO
    level = logging.WARNING
    level = logging.DEBUG
    logging.basicConfig(level=level, format='%(asctime)-15s %(message)s')
    if len(argv) not in [0, 1]:
        print "Usage: %s [db_path]" % myname
        exit(1)
    db_path = argv[0] if len(argv) > 0 else _DB_PATH
    logging.warn("Serving at %s:%s [db:%s][pid:%d]...",
                 _SERVER_ADDR, _SERVER_PORT, db_path, os.getpid())
    open_database(db_path)
    server = gevent.server.StreamServer(
	    (_SERVER_ADDR, _SERVER_PORT), handle_request)
    server.serve_forever()

if __name__ == '__main__':
    main(sys.argv[0], sys.argv[1:])



#!/usr/bin/env python
# Copyright (c) 2012 Hung-Te Lin <piaip@csie.ntu.edu.tw>. All rights reserved.
# Use of this source code is governed by a BSD license.

"""
brcstored: Board RC (brc) storage daemon.

Provides a way to manage per-user BRC object into isolated database.
"""

# Configuration of backends. 

# (Network) Use gevent of eventlet.
USE_GEVENT = True

# (Database) Use KyotoCabinet or LevelDB.
USE_KYOTO = False

#-----------------------------------------------------------------------

if USE_GEVENT:
    from gevent import monkey; monkey.patch_all()
    import gevent
    import gevent.server
else:
    import eventlet

if USE_KYOTO:
    import kyotocabinet
else:
    import leveldb

import StringIO
import logging
import os
import re
import struct
import sys
import time


_SERVER_ADDR = '127.0.0.1'
_SERVER_PORT = 5133
_DB_PATH = '/home/bbs/brcstore/database'


# Performance counters

class OperationPerf(object):

    def __init__(self, _type, _timeout=0.5):
        self._type = _type
        self._timeout = _timeout

    def __enter__(self):
        self._clock = time.time()

    def __exit__(self, *args, **kargs):
        delta = time.time() - self._clock
        if delta > self._timeout:
            sys.stderr.write("[%s %.3f]" % (self._type, delta))


class Perf(object):

    def __init__(self):
        self.read = 0
        self.write = 0
        self.req = 0
        self.clock = 0

    def add_req(self):
        if self.clock == 0:
            self.clock = time.time()
        self.req += 1

    def report(self):
        if self.req % 100 == 0:
            sys.stderr.write(
                    '+' if self.read > self.write else
                    '-' if self.read < self.write else
                    '=')
            self.read = self.write = 0
        if self.req % 7000 == 0:
            sys.stderr.write("[%.2f req/s]%s\n" %
                    (self.req / (time.time() - self.clock),
                    time.strftime('%H:%M')))
            self.clock = time.time()
            self.req = 0

    def add_read(self):
        self.read += 1

    def add_write(self):
        self.write += 1


g_perf = Perf()


def open_database(db_path):
    global g_db

    if USE_KYOTO:
        g_db = kyotocabinet.DB()
        if not g_db.open(db_path,
                         kyotocabinet.DB.OWRITER | kyotocabinet.DB.OCREATE):
            sys.stderr.write("open error: " + str(g_db.error()))
            sys.exit(1)
    else:
        # BRCv3 max size = 49152 (8192*3*2), so let's increase block size.
        # LevelDB default I/O buffer size: R=8M, W=2M.
        db = leveldb.LevelDB(db_path,
                #block_size=49152,
                # block_cache_size=(16 * (2 << 20)),
                write_buffer_size=(32 * (2 << 20)),
                )

        class LevelDBWrapper(object):

            def __init__(self, db):
                self.db = db

            def get(self, key):
                try:
                    return self.db.Get(key)
                except KeyError:
                    return None

            def set(self, key, value):
                self.db.Put(key, value)

        g_db = LevelDBWrapper(db)


def handle_request(socket, _):
    # WRITE: 'w' + UID + '\n' + int32_t len, BYTE[len]
    #  Returns: NUL
    # READ: 'r' + UID + '\n'
    #  Returns: int32_t len, BYTE[len]   (len=-1 if UID does not exist)
    fd = socket.makefile('rw')
    fmt_len = '@i'
    g_perf.add_req()
    g_perf.report()
    try:
        command = fd.read(1)
        uid = fd.readline()
        uid = uid.strip()
        if command == 'r':
            g_perf.add_read()
            with OperationPerf('R'):
                msg = g_db.get(uid)
            if msg is None:
                fd.write(struct.pack(fmt_len, -1))
                logging.info('Read : %s: (NOT FOUND)', uid)
            else:
                fd.write(struct.pack(fmt_len, len(msg)))
                fd.write(msg)
                logging.info('Read : %s: size=%d', uid, len(msg))
        elif command == 'w':
            g_perf.add_write()
            msglen = struct.unpack(fmt_len,
                                   fd.read(struct.calcsize(fmt_len)))[0]
            msg = fd.read(msglen)
	    if len(msg) != msglen:
		logging.warn('Write: incomplete: %d/%d, %s',
			len(msg), msglen, uid)
            logging.info('Write: %s: size=%d', uid, len(msg))
            with OperationPerf('W'):
                g_db.set(uid, msg)
            # Sync byte, anything should work.
            fd.write(chr(0))
        elif not command:
            raise ValueError('Incomplete request (no command).')
        else:
            raise ValueError('Unknown request: 0x%02X' % ord(command))
    except:
        logging.exception("handle_request")
    finally:
        try:
            fd.close()
            socket.close()
        except:
            pass


def main(myname, argv):
    level = logging.WARNING
    # level = logging.INFO
    # level = logging.DEBUG
    logging.basicConfig(level=level, format='%(asctime)-15s %(message)s')
    if len(argv) not in [0, 1]:
        print "Usage: %s [db_path]" % myname
        exit(1)
    db_path = argv[0] if len(argv) > 0 else _DB_PATH
    if USE_KYOTO:
        db_path += '.kch'
    logging.warn("Serving via %s at %s:%s [db:%s:%s][pid:%d]...",
                 "gevent" if USE_GEVENT else "eventlet",
                 _SERVER_ADDR, _SERVER_PORT, 
                 'KyotoCabinet' if USE_KYOTO else 'LevelDB',
                 db_path, os.getpid())
    open_database(db_path)

    if USE_GEVENT:
        server = gevent.server.StreamServer(
                (_SERVER_ADDR, _SERVER_PORT), handle_request)
        server.serve_forever()
    else:
        server = eventlet.listen((_SERVER_ADDR, _SERVER_PORT))
        pool = eventlet.GreenPool()
        while True:
            try:
                new_sock, address = server.accept()
                pool.spawn_n(handle_request, new_sock, address)
            except (SystemExit, KeyboardInterrupt):
                break


if __name__ == '__main__':
    main(sys.argv[0], sys.argv[1:])



#!/usr/bin/env python
# -*- coding: Big5 -*-

import collections
import struct

IDLEN = 12
IPV4LEN = 15
PASSLEN = 14
REGLEN = 38

PASSWD_VERSION = 4194

USEREC_SIZE = 512
USEREC_FMT = (
	("version", "I"),
	("userid", "%ds" % (IDLEN+1)),
	("realname", "20s"),
	("nickname", "24s"),
	("passwd", "%ds" % PASSLEN),
	("pad_1", "B"),
	("uflag", "I"),
	("deprecated_uflag2", "I"),
	("userlevel", "I"),
	("numlogindays", "I"),
	("numposts", "I"),
	("firstlogin", "I"),
	("lastlogin", "I"),
	("lasthost", "%ds" % (IPV4LEN+1)),
	("money", "I"),
	("_unused", "4s"),
	("email", "50s"),
	("address", "50s"),
	("justify", "%ds" % (REGLEN + 1)),
	("month", "B"),
	("day", "B"),
	("year", "B"),
	("_unused3", "B"),
	("pager_ui_type", "B"),
	("pager", "B"),
	("invisible", "B"),
	("_unused4", "2s"),
	("exmailbox", "I"),
	("_unused5", "4s"),
	("career", "40s"),
	("phone", "20s"),
	("_unused6", "I"),
	("chkpad1", "44s"),
	("role", "I"),
	("lastseen", "I"),
	("timesetangel", "I"),
	("timeplayangel", "I"),
	("lastsong", "I"),
	("loginview", "I"),
	("_unused8", "B"),
	("pad_2", "B"),
	("vl_count", "H"),
	("five_win", "H"),
	("five_lose", "H"),
	("five_tie", "H"),
	("chc_win", "H"),
	("chc_lose", "H"),
	("chc_tie", "H"),
	("mobile", "I"),
	("mind", "4s"),
	("go_win", "H"),
	("go_lose", "H"),
	("go_tie", "H"),
	("dark_win", "H"),
	("dark_lose", "H"),
	("_unused9", "B"),
	("signature", "B"),
	("_unused19", "B"),
	("badpost", "B"),
	("dark_tie", "H"),
	("myangel", "%ds" % (IDLEN + 1)),
	("pad_3", "B"),
	("chess_elo_rating", "H"),
	("withme", "I"),
	("timeremovebadpost", "I"),
	("timeviolatelaw", "I"),
	("pad_trail", "28s"),
	)

BTLEN = 48

BOARDHEADER_SIZE = 256
BOARDHEADER_FMT = (
    ("brdname", "%ds" % (IDLEN + 1)),
    ("title", "%ds" % (BTLEN + 1)),
    ("BM", "%ds" % (IDLEN * 3 + 3)),
    ("pad1", "3s"),
    ("brdattr", "I"),  # uint32_t
    ("chesscountry", "B"),
    ("vote_limit_posts", "B"),
    ("vote_limit_logins", "B"),
    ("pad2_1", "B"),
    ("bupdate", "I"),  # time4_t
    ("_post_limit_posts", "B"),
    ("post_limit_logins", "B"),
    ("pad2_2", "B"),
    ("bvote", "B"),
    ("vtime", "I"),  # time4_t
    ("level", "I"),  # uint32_t
    ("perm_reload", "I"),  # time4_t
    ("gid", "I"),  # uint32_t
    ("next0", "I"),  # uint32_t
    ("next1", "I"),  # uint32_t
    ("firstchild0", "I"),  # uint32_t
    ("firstchild1", "I"),  # uint32_t
    ("parent", "I"),  # uint32_t
    ("childcount", "I"),  # uint32_t
    ("nuser", "I"),  # uint32_t
    ("postexpire", "I"),  # uint32_t
    ("endgamble", "I"),  # time4_t
    ("posttype", "33s"),
    ("posttype_f", "B"),
    ("fastrecommend_pause", "B"),
    ("vote_limit_badpost", "B"),
    ("post_limit_badpost", "B"),
    ("pad3", "3s"),
    ("SRexpire", "I"),  # time4_t
    ("pad4", "40s"),
    )

BRD_NOCOUNT            = 0x00000002      # 不列入統計 
BRD_NOTRAN             = 0x00000004      # 不轉信 
BRD_GROUPBOARD         = 0x00000008      # 群組板 
BRD_HIDE               = 0x00000010      # 隱藏板 (看板好友才可看) 
BRD_POSTMASK           = 0x00000020      # 限制發表或閱讀 
BRD_ANONYMOUS          = 0x00000040      # 匿名板 
BRD_DEFAULTANONYMOUS   = 0x00000080      # 預設匿名板 
BRD_NOCREDIT           = 0x00000100      # 發文無獎勵看板 
BRD_VOTEBOARD          = 0x00000200      # 連署機看板 
BRD_WARNEL             = 0x00000400      # 連署機看板 
BRD_TOP                = 0x00000800      # 熱門看板群組 
BRD_NORECOMMEND        = 0x00001000      # 不可推薦 
BRD_BLOG               = 0x00002000      # (已停用) 部落格 
BRD_BMCOUNT            = 0x00004000      # 板主設定列入記錄 
BRD_SYMBOLIC           = 0x00008000      # symbolic link to board 
BRD_NOBOO              = 0x00010000      # 不可噓 
BRD_LOCALSAVE          = 0x00020000      # 預設 Local Save 
BRD_RESTRICTEDPOST     = 0x00040000      # 板友才能發文 
BRD_GUESTPOST          = 0x00080000      # guest能 post 
BRD_COOLDOWN           = 0x00100000      # 冷靜 
BRD_CPLOG              = 0x00200000      # 自動留轉錄記錄 
BRD_NOFASTRECMD        = 0x00400000      # 禁止快速推文 
BRD_IPLOGRECMD         = 0x00800000      # 推文記錄 IP 
BRD_OVER18             = 0x01000000      # 十八禁 
BRD_NOREPLY            = 0x02000000      # 不可回文 
BRD_ALIGNEDCMT         = 0x04000000      # 對齊式的推文 
BRD_NOSELFDELPOST      = 0x08000000      # 不可自刪 


FNLEN = 28  # Length of filename
TTLEN = 64  # Length of title
DISP_TTLEN = 46

FILEHEADER_FMT = (
        ("filename", "%ds" % FNLEN),
        ("modified", "I"),
        ("pad", "B"),
        ("recommend", "b"),
        ("owner", "%ds" % (IDLEN + 2)),
        ("date", "6s"),
        ("title", "%ds" % (TTLEN + 1)),
        ("pad2", "B"),
        ("multi", "i"),
        ("filemode", "B"),
        ("pad3", "3s"))

FILEHEADER_SIZE = 128

def get_format(format_pattern):
    return '<' + ''.join(value for _, value in format_pattern)

def unpack_data(blob, format_pattern):
    fmt = get_format(format_pattern)
    data = dict(zip((name for name, _ in format_pattern),
		    struct.unpack_from(fmt, blob)))
    # Convert C-style strings.
    for name, pat in format_pattern:
	if 's' in pat:
	    data[name] = data[name].partition(chr(0))[0]
    return data

FILE_LOCAL     = 0x01    # local saved,  non-mail
FILE_READ      = 0x01    # already read, mail only
FILE_MARKED    = 0x02    # non-mail + mail
FILE_DIGEST    = 0x04    # digest,       non-mail
FILE_REPLIED   = 0x04    # replied,      mail only
FILE_BOTTOM    = 0x08    # push_bottom,  non-mail
FILE_MULTI     = 0x08    # multi send,   mail only
FILE_SOLVED    = 0x10    # problem solved, sysop/BM non-mail only
FILE_HIDE      = 0x20    # hide,        in announce
FILE_BID       = 0x20    # bid,         in non-announce
FILE_BM        = 0x40    # BM only,     in announce
FILE_VOTE      = 0x40    # for vote,    in non-announce
FILE_ANONYMOUS = 0x80    # anonymous file

STRLEN = 80 # Length of most string data


if __name__ == '__main__':
    assert struct.calcsize(get_format(BOARDHEADER_FMT)) == BOARDHEADER_SIZE
    assert struct.calcsize(get_format(FILEHEADER_FMT)) == FILEHEADER_SIZE
    assert struct.calcsize(get_format(USEREC_FMT)) == USEREC_SIZE

    with open('/home/bbs/boards/A/ALLPOST/.DIR', 'rb') as f:
	entry = f.read(FILEHEADER_SIZE)
    header = unpack_data(entry, FILEHEADER_FMT)
    print header

    with open('/home/bbs/.PASSWDS', 'rb') as f:
	entry = f.read(USEREC_SIZE)
    user = unpack_data(entry, USEREC_FMT)
    print user

    with open('/home/bbs/.BRD', 'rb') as f:
	entry = f.read(BOARDHEADER_SIZE)
    board = unpack_data(entry, BOARDHEADER_FMT)
    print board

    print 'file title: ', header['title'].decode('big5').encode('big5')
    print 'user name: ', user['realname'].decode('big5').encode('big5')
    print 'board title: ', board['title'].decode('big5').encode('big5')
